@@ -134,9 +134,8 @@
 //!
 //! To write `x ~~ y` you use `q(x, y)` or `qual(x, y)`.
 //!
+//! `q(x, y)` is the same as `and!(eq(x, y), qubit(x), qubit(y))`.
 //! `q(x, x)` is the same as `qubit(x)`.
-//! `q(x, y)` where `x` and `y` are symbolic distinct is the same as
-//! `and!(eq(x, y), q(x, x), q(y, y))`.
 //!
 //! A qubit is a kind of "superposition".
 //! One can also think about it as introducing a new argument `qubit(x)` that depends on `x`.
@@ -597,8 +596,13 @@ pub fn qubit(a: u64) -> u64 {
     use rand::{Rng, SeedableRng};
     use rand::rngs::StdRng;
     let r = unsafe {&*current::Current::<u64>::new()};
-    let mut rng = StdRng::seed_from_u64(a ^ *r);
-    rng.gen()
+    if a & 1 == 1 {
+        let mut rng = StdRng::seed_from_u64(not(a) ^ *r);
+        not(rng.gen())
+    } else {
+        let mut rng = StdRng::seed_from_u64(a ^ *r);
+        rng.gen()
+    }
 }
 
 /// Amplify a "wavefunction" of a proposition using its qubit transform.
@@ -610,8 +614,17 @@ pub fn amplify(n: u32, mut a: u64) -> u64 {
 }
 
 /// Path semantical quality `a ~~ b`.
-pub fn qual(a: u64, b: u64) -> u64 {
-    and(eq(a, b), if a == b {qubit(a)} else {and(qual(a, a), qual(b, b))})
+pub fn qual(a: u64, b: u64) -> u64 {and!(eq(a, b), qubit(a), qubit(b))}
+
+/// Aligns equality of qubits up to some homotopy level.
+pub fn hom_eq(n: u32, mut a: u64, mut b: u64) -> u64 {
+    let mut res = T;
+    for _ in 0..n {
+        res = and(res, eq(a, b));
+        a = qubit(a);
+        b = qubit(b);
+    }
+    res
 }
 
 /// Assumes the path semantical core axiom.
@@ -619,6 +632,56 @@ pub fn ps_core(a: u64, b: u64, c: u64, d: u64) -> u64 {
     imply(and!(qual(a, b), imply(a, c), imply(b, d)), qual(c, d))
 }
 
+/// Defines a proposition relation of proposition `x` to potential proofs `a` and `b`.
+pub fn is_prop(x: u64, a: u64, b: u64) -> u64 {
+    imply(
+        and(imply(a, x), imply(b, x)),
+        imply(T, eq(a, b))
+    )
+}
+
+/// Defines a set relation from a set `x` to potential members `a` and `b`.
+pub fn is_set(x: u64, a: u64, b: u64) -> u64 {
+    imply(
+        and(imply(a, x), imply(b, x)),
+        imply(eq(a, b), eq(qubit(a), qubit(b)))
+    )
+}
+
+/// Defines a groupoid relation from `x` to `a` and `b`.
+pub fn is_groupoid(x: u64, a: u64, b: u64) -> u64 {
+    imply(
+        and(imply(a, x), imply(b, x)),
+        imply(eq(qubit(a), qubit(b)), eq(qubit(qubit(a)), qubit(qubit(b))))
+    )
+}
+
+/// Defines an n-groupoid relation from `x` to `a` and `b`.
+pub fn is_groupoid_n(n: u32, x: u64, a: u64, b: u64) -> u64 {
+    let cond = and(imply(a, x), imply(b, x));
+    let mut a = a;
+    let mut b = b;
+    for _ in 0..n {
+        a = qubit(a);
+        b = qubit(b);
+    }
+    imply(cond, imply(eq(a, b), eq(qubit(a), qubit(b))))
+}
+
+/// Defines a homotopy level `n` relation from `x` to `a` and `b`.
+pub fn is_hom_lev_n(n: u32, x: u64, a: u64, b: u64) -> u64 {
+    match n {
+        0 => T,
+        1 => is_prop(x, a, b),
+        n => is_groupoid_n(n - 2, x, a, b),
+    }
+}
+
+/// Assumes univalence axiom for some homotopy level.
+pub fn univ(n: u32, a: u64, b: u64) -> u64 {
+    hom_eq(n, eq(a, b), hom_eq(n, a, b))
+}
+
 /// Measures result repeatedly.
 pub fn measure<O: Observable>(n: u32, mut fun: impl FnMut() -> O) -> O {
     let mut b = O::max_energy();
